﻿Imports SeedEditor.Seed
Imports CsGL.OpenGL

Public Class Form1

    Dim AnimFile As AnimFile
    Dim WithEvents myGLView As GLModelView

#Region "Mouse handling"
    'Camera movement
    Dim InitialMouseX, InitialMouseY As Integer ' When pressed down
    Dim PrevMouseX, PrevMouseY As Integer ' In previous frame
    Dim CameraMoveMode As CameraMovement = CameraMovement.None
    Dim RotationSpeed As Double = 0.18
    Dim MoveSpeed As Double = 4
    Dim GridAccumulator As Single

    Enum CameraMovement
        None
        Left
        Right
        LeftRight
    End Enum

    Sub SetCamModeFromMousePress(btn As MouseButtons)
        If btn = Windows.Forms.MouseButtons.Left Then
            If CameraMoveMode = CameraMovement.Right Then
                CameraMoveMode = CameraMovement.LeftRight
            Else
                CameraMoveMode = CameraMovement.Left
            End If
        ElseIf btn = Windows.Forms.MouseButtons.Right Then
            If CameraMoveMode = CameraMovement.Left Then
                CameraMoveMode = CameraMovement.LeftRight
            Else
                CameraMoveMode = CameraMovement.Right
            End If
        ElseIf btn = (Windows.Forms.MouseButtons.Left Or Windows.Forms.MouseButtons.Right) Then
            CameraMoveMode = CameraMovement.LeftRight
        ElseIf btn = Windows.Forms.MouseButtons.None Then
            CameraMoveMode = CameraMovement.None
        End If
    End Sub

    Sub SetCamModeFromMouseRelease(btn As MouseButtons)
        If CameraMoveMode = CameraMovement.LeftRight Then
            If btn = Windows.Forms.MouseButtons.Left Then
                CameraMoveMode = CameraMovement.Right
            ElseIf btn = Windows.Forms.MouseButtons.Right Then
                CameraMoveMode = CameraMovement.Left
            End If
        Else
            CameraMoveMode = CameraMovement.None
        End If
    End Sub

    Sub WrapMouse()
        Dim pos = MousePosInPreview()
        Dim newpos = pos

        If pos.X >= myGLView.Width Then
            newpos.X -= myGLView.Width
        ElseIf pos.X < 0 Then
            newpos.X += myGLView.Width
        End If

        If pos.Y >= myGLView.Height Then
            newpos.Y -= myGLView.Height
        ElseIf pos.Y < 0 Then
            newpos.Y += myGLView.Height
        End If

        PrevMouseX += newpos.X - pos.X
        PrevMouseY += newpos.Y - pos.Y

        SetMousePosition(newpos)
    End Sub

    Sub SetMousePosition(pos As Point)
        Cursor.Position = myGLView.PointToScreen(pos)
    End Sub

    Function MousePosInPreview() As Point
        Return myGLView.PointToClient(Cursor.Position)
    End Function

    Function DegToRad(angle As Double) As Double
        Return angle / 360 * 2 * Math.PI
    End Function

    Private Sub myGLView_MouseDown(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseDown
        InitialMouseX = e.X
        InitialMouseY = e.Y
        PrevMouseX = e.X
        PrevMouseY = e.Y
        GridAccumulator = 0
        SetCamModeFromMousePress(e.Button)

        UpdateView()
    End Sub

    Private Sub myGLView_MouseUp(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseUp
        SetCamModeFromMouseRelease(e.Button)
    End Sub

    Private Sub myGLView_MouseMove(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseMove
        Dim MouseFrameDeltaX, MouseFrameDeltaY As Integer

        'in most cases it is unnecessary
        If e.Button = Windows.Forms.MouseButtons.None Then
            CameraMoveMode = CameraMovement.None
        End If

        If CameraMoveMode <> CameraMovement.None Then
            WrapMouse()
        End If

        MouseFrameDeltaX = e.X - PrevMouseX
        MouseFrameDeltaY = e.Y - PrevMouseY

        Select Case CameraMoveMode
            Case CameraMovement.Right
                myGLView.CameraRotationH += MouseFrameDeltaX * RotationSpeed
                myGLView.CameraRotationV += MouseFrameDeltaY * RotationSpeed
                myGLView.CameraRotationV = Math.Max(-90, Math.Min(90, myGLView.CameraRotationV))

                UpdateView()
            Case CameraMovement.Left
                myGLView.CameraRotationH += MouseFrameDeltaX * RotationSpeed

                Dim moveDelta As Vector
                moveDelta.X = Math.Cos(DegToRad(-myGLView.CameraRotationH + 90))
                moveDelta.Z = Math.Sin(DegToRad(-myGLView.CameraRotationH + 90))
                moveDelta *= -MouseFrameDeltaY
                moveDelta *= MoveSpeed

                myGLView.CameraPosition += moveDelta
                UpdateView()
            Case CameraMovement.LeftRight
                myGLView.CameraPosition.Y -= MouseFrameDeltaY * MoveSpeed

                Dim moveDelta As Vector
                moveDelta.X = Math.Cos(DegToRad(-myGLView.CameraRotationH))
                moveDelta.Z = Math.Sin(DegToRad(-myGLView.CameraRotationH))
                moveDelta *= MouseFrameDeltaX
                moveDelta *= MoveSpeed

                myGLView.CameraPosition += moveDelta

                UpdateView()
        End Select

        PrevMouseX = e.X
        PrevMouseY = e.Y
    End Sub
#End Region

    Private Sub UpdateView()
        myGLView.GLDraw()
        myGLView.Refresh()
    End Sub

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If OpenAnimDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            AnimFile = New AnimFile(OpenAnimDlg.FileName)
            Text = AnimFile.Name & " - SeedAnimReader"
            myGLView.AnimFile = AnimFile

            tvBones.BeginUpdate()
            tvBones.Nodes.Clear()
            For Each bone In AnimFile.Bones
                If bone.Parent Is Nothing Then
                    Dim node = tvBones.Nodes.Add(bone.Name)
                    node.Tag = bone
                    AddChildBones(node)
                End If
            Next
            tvBones.ExpandAll()
            tvBones.EndUpdate()

            lvAnimations.BeginUpdate()
            lvAnimations.Items.Clear()
            For Each seq In AnimFile.Sequences
                lvAnimations.Items.Add(seq.Name).SubItems.Add(seq.Frames.Count)
            Next
            If lvAnimations.Items.Count > 0 Then
                lvAnimations.Items(0).Selected = True
            End If
            lvAnimations.EndUpdate()

            SetFrame(0, 0)
        End If

    End Sub

    Sub AddChildBones(node As TreeNode)
        Dim bone As AnimBone = node.Tag
        For Each child In bone.Children
            Dim subNode = node.Nodes.Add(child.Name)
            subNode.Tag = child
            AddChildBones(subNode)
        Next
    End Sub

    Private Sub Form1_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        myGLView = New GLModelView
        myGLView.Dock = DockStyle.Fill
        myGLView.Parent = ViewportPanel
    End Sub

    Private Sub lvAnimations_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lvAnimations.SelectedIndexChanged
        If lvAnimations.SelectedIndices.Count = 1 Then
            Dim iSequence = lvAnimations.SelectedIndices(0)
            tbFrame.Value = 0
            tbFrame.Maximum = AnimFile.Sequences(iSequence).Frames.Count - 1
            SetFrame(iSequence, 0)
            UpdateView()
        End If
    End Sub

    Private Sub tbFrame_Scroll(sender As System.Object, e As System.EventArgs) Handles tbFrame.Scroll
        If lvAnimations.SelectedIndices.Count = 1 Then
            SetFrame(lvAnimations.SelectedIndices(0), tbFrame.Value)
        End If
    End Sub

    Sub SetFrame(iSequence As Integer, iFrame As Integer)
        Dim frame = AnimFile.Sequences(iSequence).Frames(iFrame)

        For i = 0 To AnimFile.Bones.Count - 1
            Dim bone = AnimFile.Bones(i)
            bone.CurMatrix = frame.BonePositions(i)
        Next

        For Each bone In AnimFile.Bones
            If bone.Parent Is Nothing Then
                SetUpChildBones(bone)
            End If
        Next

        UpdateView()
    End Sub

    Sub SetUpChildBones(parent As AnimBone)
        For Each child In parent.Children
            child.CurMatrix = parent.CurMatrix * child.CurMatrix
            SetUpChildBones(child)
        Next
    End Sub

    Private Sub tvBones_AfterSelect(sender As System.Object, e As System.Windows.Forms.TreeViewEventArgs) Handles tvBones.AfterSelect
        For Each bone In AnimFile.Bones
            If tvBones.SelectedNode IsNot Nothing Then
                bone.Selected = (bone Is tvBones.SelectedNode.Tag)
            Else
                bone.Selected = False
            End If
        Next
        UpdateView()
    End Sub
End Class

Public Class GLModelView
    Inherits OpenGLControl

    Public AnimFile As AnimFile

    Public CameraPosition As Vector = New Vector(0, 128, -512)
    Public CameraRotationH As Single
    Public CameraRotationV As Single

    Protected Overrides Sub initGLContext() 'Need to override this method
        GL.glEnable(GLFlags.GL_CULL_FACE) ' dont draw back
        GL.glShadeModel(GLFlags.GL_SMOOTH) 'Smooth Shading
        GL.glClearColor(0.0F, 0.0F, 0.0, 0.5F)               'Black Background
        GL.glClearDepth(1.0F)                                'Depth Buffer Setup
        GL.glEnable(GLFlags.GL_DEPTH_TEST) 'Enable Depth Testing
        GL.glDepthFunc(GLFlags.GL_LEQUAL)  'Type of Depth Testing to do
        'does this do anything?
        GL.glHint(GLFlags.GL_PERSPECTIVE_CORRECTION_HINT, GLFlags.GL_NICEST) 'Nice Perspective Calculations
    End Sub

    Protected Overrides Sub OnSizeChanged(ByVal e As EventArgs)
        MyBase.OnSizeChanged(e)
        Dim S As Size = Me.Size

        GL.glMatrixMode(GLFlags.GL_PROJECTION)
        GL.glLoadIdentity()
        GL.gluPerspective(90, S.Width / S.Height, 1, 65536.0)
        GL.glMatrixMode(GLFlags.GL_MODELVIEW)
        GL.glLoadIdentity()
    End Sub

    Private Sub DrawLine3D(x1 As Single, y1 As Single, z1 As Single, x2 As Single, y2 As Single, z2 As Single, col As Color)
        GL.glBegin(GLFlags.GL_LINES)
        GL.glColor3ub(col.R, col.G, col.B)
        GL.glVertex3f(x1, y1, z1)
        GL.glVertex3f(x2, y2, z2)
        GL.glEnd()
    End Sub

    Private Sub DrawLine3D(v1 As Vector, v2 As Vector, col As Color)
        DrawLine3D(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, col)
    End Sub

    Private Sub DrawBackgroundGrid()
        Dim normal = Color.FromArgb(64, 64, 64)
        Dim dark = Color.FromArgb(48, 48, 48)

        Dim a, b As Vector

        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        GL.glColor3b(32, 32, 32)
        GL.glutWireCube(65536)

        For i = -32768 To 32768 Step 1024
            a.X = i
            b.X = i
            a.Z = -32768
            b.Z = 32768
            DrawLine3D(a, b, If(i = 0, normal, dark))
            a.X = -32768
            b.X = 32768
            a.Z = i
            b.Z = i
            DrawLine3D(a, b, If(i = 0, normal, dark))
        Next
    End Sub

    Public Overrides Sub GLDraw() 'We use this method to draw our objects on screen
        GL.glDisable(GL.GL_BLEND)
        GL.glDepthMask(GL.GL_TRUE)

        GL.glEnable(GL.GL_TEXTURE_2D)

        GL.glClearColor(0, 0, 0, 1)
        GL.glClear(GLFlags.GL_COLOR_BUFFER_BIT Or GLFlags.GL_DEPTH_BUFFER_BIT)

        GL.glMatrixMode(GLFlags.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMultMatrixf({1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1}) 'remap coord system
        GL.glRotatef(-CameraRotationV, 1, 0, 0) 'up/down
        GL.glRotatef(-CameraRotationH, 0, 1, 0) 'left/right
        GL.glTranslatef(-CameraPosition.X, -CameraPosition.Y, -CameraPosition.Z)
        GL.glFrontFace(GL.GL_CW) 'Yup, now we need that

        DrawBackgroundGrid()
        GL.glClear(GLFlags.GL_DEPTH_BUFFER_BIT)
        GL.glEnable(GL.GL_TEXTURE_2D)

        If AnimFile IsNot Nothing Then
            For i = 0 To AnimFile.Bones.Count - 1
                Dim bone = AnimFile.Bones(i)
                Dim boneMtx = bone.CurMatrix
                If bone.Parent IsNot Nothing Then
                    Dim parentMtx = bone.Parent.CurMatrix
                    DrawLine3D(boneMtx.Xfrom1, boneMtx.Yfrom1, boneMtx.Zfrom1,
                               parentMtx.Xfrom1, parentMtx.Yfrom1, parentMtx.Zfrom1, Color.White)
                End If
                Dim brightness = 128
                If bone.Selected Then brightness = 255
                DrawLine3D(New Vector(boneMtx.Xfrom1, boneMtx.Yfrom1, boneMtx.Zfrom1), New Vector(16, 0, 0).TransformBy(boneMtx), Color.FromArgb(brightness, 0, 0))
                DrawLine3D(New Vector(boneMtx.Xfrom1, boneMtx.Yfrom1, boneMtx.Zfrom1), New Vector(0, 16, 0).TransformBy(boneMtx), Color.FromArgb(0, brightness, 0))
                DrawLine3D(New Vector(boneMtx.Xfrom1, boneMtx.Yfrom1, boneMtx.Zfrom1), New Vector(0, 0, 16).TransformBy(boneMtx), Color.FromArgb(0, 0, brightness))
            Next
        End If

        GL.glFlush()
    End Sub

End Class