﻿Public Class Form1

    Dim curProject As Project
    Dim temporaryFiles As New List(Of String)
    Dim temporaryFileHandles As New List(Of IO.FileStream)
    Dim hexEditorPath As String = ""

    Private Sub btnOpen_Click(sender As System.Object, e As System.EventArgs) Handles btnOpen.Click
        If curProject IsNot Nothing Then
            OpenDlg.InitialDirectory = IO.Path.GetDirectoryName(curProject.File)
        End If
        If OpenDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            LoadFile(OpenDlg.FileName)
        End If
    End Sub

    Sub LoadFile(file As String)
        curProject = New Project(file)

        lvAssets.BeginUpdate()
        lvAssets.Items.Clear()
        For Each asset In curProject.Assets
            Dim item = lvAssets.Items.Add(asset.OriginalOffset)
            item.SubItems.Add(asset.OriginalSize)
            item.SubItems.Add(asset.ID)
            item.SubItems.Add(asset.Type.ToString)
            item.SubItems.Add(asset.Name)
        Next
        lvAssets.EndUpdate()

        Me.Text = IO.Path.GetFileName(file) & " - SeedEditor"
    End Sub

    Private Sub btnExportAssets_Click(sender As System.Object, e As System.EventArgs) Handles btnExportAssets.Click
        If curProject Is Nothing Then Return

        If ExportDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            Dim rootDir = ExportDlg.SelectedPath

            For Each a In curProject.Assets
                If a.Type = AssetType.Texture Then
                    DirectCast(a, Texture).Bitmap.Save(IO.Path.Combine(rootDir, a.Name & ".bmp"), Imaging.ImageFormat.Bmp)
                ElseIf a.Type = AssetType.Model Then
                    DirectCast(a, Model).SaveOBJ(IO.Path.Combine(rootDir, a.Name & ".obj"))
                End If
            Next
            MessageBox.Show("Done")
        End If
    End Sub

    Private Sub llAbout_LinkClicked(sender As System.Object, e As System.Windows.Forms.LinkLabelLinkClickedEventArgs) Handles llAbout.LinkClicked
        MessageBox.Show(My.Resources.About, "About")
    End Sub

    Private Sub btnSave_Click(sender As System.Object, e As System.EventArgs) Handles btnSave.Click
        If curProject Is Nothing Then Return

        SaveDlg.FileName = curProject.File
        SaveDlg.InitialDirectory = IO.Path.GetDirectoryName(curProject.File)
        If SaveDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            If IO.File.Exists(SaveDlg.FileName) Then
                Dim backupName = SaveDlg.FileName & "_orig"
                If Not IO.File.Exists(backupName) Then
                    IO.File.Move(SaveDlg.FileName, backupName)
                End If
            End If
            curProject.Save(SaveDlg.FileName)
            Me.Text = IO.Path.GetFileName(SaveDlg.FileName) & " - SeedReader"
        End If
    End Sub

    Private Sub Form1_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        If IO.File.Exists("SeedReader.ini") Then
            Using r As New IO.StreamReader("SeedReader.ini")
                While Not r.EndOfStream
                    Dim line = r.ReadLine
                    Dim equalsPos = line.IndexOf("=")
                    Dim prop = line.Substring(0, equalsPos)
                    Dim value = line.Substring(equalsPos + 1)
                    Select Case prop
                        Case "hex_editor_path"
                            hexEditorPath = value
                        Case "textures_path"
                            OpenBmpDlg.InitialDirectory = value
                        Case "models_path"
                            OpenObjDlg.InitialDirectory = value
                    End Select
                End While
            End Using
        End If

        If My.Application.CommandLineArgs.Count = 1 Then
            LoadFile(My.Application.CommandLineArgs(0))
        End If
    End Sub

    Private Sub btnReplaceAsset_Click(sender As System.Object, e As System.EventArgs) Handles btnReplaceAsset.Click
        If lvAssets.SelectedItems.Count = 1 Then
            Dim asset = curProject.Assets(lvAssets.SelectedIndices(0))

            Select Case asset.Type
                Case AssetType.Texture
                    If OpenBmpDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
                        OpenBmpDlg.InitialDirectory = IO.Path.GetDirectoryName(OpenBmpDlg.FileName)

                        Dim bmp = New Bitmap(OpenBmpDlg.FileName)
                        If IsPowerOfTwo(bmp.Width) And IsPowerOfTwo(bmp.Height) Then
                            DirectCast(asset, Texture).Bitmap = bmp
                        Else
                            MessageBox.Show("Width and height must be power of two.")
                        End If
                    End If
                Case AssetType.Model
                    If OpenObjDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
                        OpenObjDlg.InitialDirectory = IO.Path.GetDirectoryName(OpenObjDlg.FileName)
                        DirectCast(asset, Model).LoadOBJ(OpenObjDlg.FileName)
                    End If
            End Select
        End If
    End Sub

    Private Sub tbSearch_TextChanged(sender As System.Object, e As System.EventArgs) Handles tbSearch.TextChanged
        Dim text = tbSearch.Text
        If text <> "" Then
            For i = 0 To curProject.Assets.Count - 1
                If curProject.Assets(i).Name.StartsWith(text, StringComparison.OrdinalIgnoreCase) Then
                    lvAssets.Items(i).Selected = True
                    lvAssets.EnsureVisible(i)
                    Return
                End If
            Next
        End If
    End Sub

    Private Sub lvAssets_DoubleClick(sender As System.Object, e As System.EventArgs) Handles lvAssets.DoubleClick
        If lvAssets.SelectedItems.Count = 1 Then
            Dim asset = curProject.Assets(lvAssets.SelectedIndices(0))

            If 1 Then
            End If
        End If
    End Sub

    Private Sub Form1_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        Using w As New IO.StreamWriter("SeedReader.ini.tmp")
            w.WriteLine("hex_editor_path=" & hexEditorPath)
            w.WriteLine("textures_path=" & OpenBmpDlg.InitialDirectory)
            w.WriteLine("models_path=" & OpenObjDlg.InitialDirectory)
        End Using

        IO.File.Delete("SeedReader.ini")
        IO.File.Move("SeedReader.ini.tmp", "SeedReader.ini")

        For i = 0 To temporaryFiles.Count - 1
            temporaryFileHandles(i).Close()
            Try
                IO.File.Delete(temporaryFiles(i))
            Catch
            End Try
        Next
    End Sub

    Private Sub btnEditAsset_Click(sender As System.Object, e As System.EventArgs) Handles btnEditAsset.Click
        If lvAssets.SelectedItems.Count = 1 Then
            Dim asset = curProject.Assets(lvAssets.SelectedIndices(0))

            Select Case asset.Type
                Case AssetType.Light
                    LightProps.Light = asset
                    LightProps.ShowDialog()
                Case AssetType.Model
                    ModelProps.Model = asset
                    ModelProps.ShowDialog()
                Case AssetType.ActiveA, AssetType.ActiveB
                    Dim propsForm As New ActiveProps
                    propsForm.Active = asset
                    propsForm.ShowDialog(Me)
                    Return
            End Select
        End If
    End Sub

    Private Sub lvMain_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lvAssets.SelectedIndexChanged
        btnEditAsset.Enabled = False
        btnReplaceAsset.Enabled = False
        If lvAssets.SelectedItems.Count = 1 Then
            Dim asset = curProject.Assets(lvAssets.SelectedIndices(0))

            Select Case asset.Type
                Case AssetType.Light, AssetType.ActiveA, AssetType.ActiveB
                    btnEditAsset.Enabled = True
                Case AssetType.Model
                    btnEditAsset.Enabled = True
                    btnReplaceAsset.Enabled = True
                Case AssetType.Texture
                    btnReplaceAsset.Enabled = True
            End Select
        End If
    End Sub

    Private Sub btn3DView_Click(sender As System.Object, e As System.EventArgs) Handles btn3DEditor.Click
        If curProject Is Nothing Then Return

        ThreeDEditorForm.Project = curProject
        ThreeDEditorForm.ShowDialog(Me)
    End Sub

    Private Sub lvAssets_DragEnter(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles lvAssets.DragEnter
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim files() As String = e.Data.GetData(DataFormats.FileDrop)

            Select Case LCase(IO.Path.GetExtension(files(0)))
                Case ".seed"
                    e.Effect = DragDropEffects.All
            End Select
        End If
    End Sub

    Private Sub lvAssets_DragDrop(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles lvAssets.DragDrop
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim files() As String = e.Data.GetData(DataFormats.FileDrop)
            Select LCase(IO.Path.GetExtension(files(0)))
                Case ".seed"
                    LoadFile(files(0))
            End Select
        End If
    End Sub

    Private Sub btnSaveB_Click(sender As System.Object, e As System.EventArgs) Handles btnSaveB.Click
        Dim exporter As New CustomActiveExporter
        Dim items As New List(Of ActiveB)
        For Each a In curProject.Assets
            If TypeOf a Is ActiveB Then
                Dim ab = DirectCast(a, ActiveB)
                If ab.Parent Is Nothing Then
                    items.Add(ab)
                End If
            End If
        Next
        exporter.Export(items, "Custom actives")
    End Sub
End Class

