﻿Public Class ModelProps
    Public Model As Model
    Dim ShownTextures As New List(Of Texture)
    Dim ChangingShownTexture As Boolean

    Private Sub ModelProps_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        Text = "Model $" & Model.ID & " " & Model.Name
        cbDiffuse.Checked = False
        cbEmissive.Checked = False
        rbGroupNoChange.Checked = True
        nuRoomID.Value = Model.Room.ID
        nuMisc1.Value = Model.Unknown1
        nuMisc2.Value = Model.Unknown3
        nuTransparency.Value = Model.Transparency
        If Model.Polygons.Count > 0 Then
            nuDiffuse.Value = Model.Polygons(0).Diffuse
            nuEmissiveR.Value = Model.Polygons(0).EmissiveColor.X
            nuEmissiveG.Value = Model.Polygons(0).EmissiveColor.Y
            nuEmissiveB.Value = Model.Polygons(0).EmissiveColor.Z
        End If
        For i = 0 To 32 - 1
            clbFlags.SetItemChecked(i, (Model.Flags >> i) And 1)
        Next
        Dim prj = Model.Project
        lbTextures.Items.Clear()
        For Each t In Model.Textures
            lbTextures.Items.Add(t.Name)
        Next
        cmbTexture.Items.Clear()
        ShownTextures.Clear()
        For Each a In prj.Assets
            If a.Type = AssetType.Texture Then
                cmbTexture.Items.Add(a.ID & ": " & a.Name)
                ShownTextures.Add(a)
            End If
        Next
    End Sub

    Private Sub ModelProps_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        Model.Room = Model.Project.GetAsset(nuRoomID.Value)
        Model.Unknown1 = nuMisc1.Value
        Model.Unknown3 = nuMisc2.Value
        Model.Transparency = nuTransparency.Value
        If cbDiffuse.Checked Or cbEmissive.Checked Then
            For i = 0 To Model.Polygons.Count - 1
                Dim poly = Model.Polygons(i)
                If cbDiffuse.Checked Then
                    poly.Diffuse = nuDiffuse.Value
                End If
                If cbEmissive.Checked Then
                    poly.EmissiveColor.X = nuEmissiveR.Value
                    poly.EmissiveColor.Y = nuEmissiveG.Value
                    poly.EmissiveColor.Z = nuEmissiveB.Value
                End If
                Model.Polygons(i) = poly
            Next
        End If

        Model.Flags = 0
        For i = 0 To 32 - 1
            If clbFlags.GetItemChecked(i) Then
                Model.Flags = Model.Flags Or (1 << i)
            End If
        Next

        If rbGroupOne.Checked Then
            Model.PolygonGroups.Clear()
            Model.PolygonGroups.Add(New PolygonGroup(0, Model.Polygons.Count))
        ElseIf rbGroupPerPoly.Checked Then
            Model.PolygonGroups.Clear()
            For i = 0 To Model.Polygons.Count - 1
                Model.PolygonGroups.Add(New PolygonGroup(i, 1))
            Next
        End If
    End Sub

    Private Sub btnEmissive_Click(sender As System.Object, e As System.EventArgs) Handles btnEmissive.Click
        Dim r, g, b As Byte
        r = Math.Max(0, Math.Min(255, nuEmissiveR.Value * 255))
        g = Math.Max(0, Math.Min(255, nuEmissiveG.Value * 255))
        b = Math.Max(0, Math.Min(255, nuEmissiveB.Value * 255))
        ColorDlg.Color = Color.FromArgb(r, g, b)

        If ColorDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            nuEmissiveR.Value = ColorDlg.Color.R / 255
            nuEmissiveG.Value = ColorDlg.Color.G / 255
            nuEmissiveB.Value = ColorDlg.Color.B / 255
        End If
    End Sub

    Private Sub cbDiffuse_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbDiffuse.CheckedChanged
        gbDiffuse.Enabled = cbDiffuse.Checked
    End Sub

    Private Sub cbEmissive_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbEmissive.CheckedChanged
        gbEmissive.Enabled = cbEmissive.Checked
    End Sub

    Private Sub lbTextures_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles lbTextures.SelectedIndexChanged
        If lbTextures.SelectedItems.Count = 1 Then
            cmbTexture.Enabled = True
            Dim iTexture = lbTextures.SelectedIndex
            Dim tex = Model.Textures(iTexture)
            ChangingShownTexture = True
            cmbTexture.SelectedIndex = ShownTextures.FindIndex(Function(x) x.ID = tex.ID)
            ChangingShownTexture = False
        Else
            cmbTexture.Enabled = False
        End If
    End Sub

    Private Sub cmbTexture_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles cmbTexture.SelectedIndexChanged
        If Not ChangingShownTexture Then
            If lbTextures.SelectedItems.Count = 1 Then
                Dim iTexture = lbTextures.SelectedIndex
                Dim texAsset = ShownTextures(cmbTexture.SelectedIndex)
                Model.Textures(iTexture) = texAsset
                lbTextures.Items(iTexture) = texAsset.Name
            End If
        End If
    End Sub

End Class