﻿Imports CsGL.OpenGL

Public Class ThreeDEditorForm

    Public Project As Project
    Public Scene As Scene
    Public CustomActives As New List(Of ActiveB)
    Dim Grid As Integer

    Dim WithEvents myGLView As GLViewControl

#Region "Mouse handling"
    'Camera movement
    Dim InitialMouseX, InitialMouseY As Integer ' When pressed down
    Dim PrevMouseX, PrevMouseY As Integer ' In previous frame
    Dim CameraMoveMode As CameraMovement = CameraMovement.None
    Dim RotationSpeed As Double = 0.18
    Dim MoveSpeed As Double = 4
    Dim GridAccumulator As Single

    Enum CameraMovement
        None
        Left
        Right
        LeftRight
    End Enum

    Sub SetCamModeFromMousePress(btn As MouseButtons)
        If btn = Windows.Forms.MouseButtons.Left Then
            If CameraMoveMode = CameraMovement.Right Then
                CameraMoveMode = CameraMovement.LeftRight
            Else
                CameraMoveMode = CameraMovement.Left
            End If
        ElseIf btn = Windows.Forms.MouseButtons.Right Then
            If CameraMoveMode = CameraMovement.Left Then
                CameraMoveMode = CameraMovement.LeftRight
            Else
                CameraMoveMode = CameraMovement.Right
            End If
        ElseIf btn = (Windows.Forms.MouseButtons.Left Or Windows.Forms.MouseButtons.Right) Then
            CameraMoveMode = CameraMovement.LeftRight
        ElseIf btn = Windows.Forms.MouseButtons.None Then
            CameraMoveMode = CameraMovement.None
        End If
    End Sub

    Sub SetCamModeFromMouseRelease(btn As MouseButtons)
        If CameraMoveMode = CameraMovement.LeftRight Then
            If btn = Windows.Forms.MouseButtons.Left Then
                CameraMoveMode = CameraMovement.Right
            ElseIf btn = Windows.Forms.MouseButtons.Right Then
                CameraMoveMode = CameraMovement.Left
            End If
        Else
            CameraMoveMode = CameraMovement.None
        End If
    End Sub

    Sub WrapMouse()
        Dim pos = MousePosInPreview()
        Dim newpos = pos

        If pos.X >= myGLView.Width Then
            newpos.X -= myGLView.Width
        ElseIf pos.X < 0 Then
            newpos.X += myGLView.Width
        End If

        If pos.Y >= myGLView.Height Then
            newpos.Y -= myGLView.Height
        ElseIf pos.Y < 0 Then
            newpos.Y += myGLView.Height
        End If

        PrevMouseX += newpos.X - pos.X
        PrevMouseY += newpos.Y - pos.Y

        SetMousePosition(newpos)
    End Sub

    Sub SetMousePosition(pos As Point)
        Cursor.Position = myGLView.PointToScreen(pos)
    End Sub

    Function MousePosInPreview() As Point
        Return myGLView.PointToClient(Cursor.Position)
    End Function

    Function DegToRad(angle As Double) As Double
        Return angle / 360 * 2 * Math.PI
    End Function

    Private Sub myGLView_MouseDown(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseDown
        InitialMouseX = e.X
        InitialMouseY = e.Y
        PrevMouseX = e.X
        PrevMouseY = e.Y
        GridAccumulator = 0
        SetCamModeFromMousePress(e.Button)

        UpdateView()
    End Sub

    Private Sub myGLView_MouseUp(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseUp
        SetCamModeFromMouseRelease(e.Button)

        ' Clicking on entities
        If e.X = InitialMouseX And e.Y = InitialMouseY Then
            Dim mousepos = myGLView.PointToClient(MousePosition)
            Dim clickedID = myGLView.GetIDUnderPoint(mousepos)

            If ModifierKeys <> Keys.Control Then
                DeselectAll()
            End If

            Dim clickedAsset = Project.GetAsset(clickedID)
            If clickedAsset IsNot Nothing Then
                clickedAsset.Selected = Not clickedAsset.Selected
            End If

            OnWindowSelectionChange()

            UpdateView()
        End If

    End Sub

    Private Sub myGLView_MouseMove(ByVal sender As Object, ByVal e As MouseEventArgs) Handles myGLView.MouseMove
        Dim MouseFrameDeltaX, MouseFrameDeltaY As Integer

        'in most cases it is unnecessary
        If e.Button = Windows.Forms.MouseButtons.None Then
            CameraMoveMode = CameraMovement.None
        End If

        If CameraMoveMode <> CameraMovement.None Then
            WrapMouse()
        End If

        MouseFrameDeltaX = e.X - PrevMouseX
        MouseFrameDeltaY = e.Y - PrevMouseY

        If ModifierKeys = Keys.Control Then ' Moving objects
            Dim moveObjectBy = 0

            If CameraMoveMode = CameraMovement.LeftRight Then
                GridAccumulator -= MouseFrameDeltaY * 1.5
            Else
                GridAccumulator += MouseFrameDeltaX * 1.5
            End If

            If Math.Abs(GridAccumulator) >= grid Then
                moveObjectBy = (GridAccumulator \ grid) * grid
                GridAccumulator -= moveObjectBy
            End If

            For Each mdl In Scene.Models
                If mdl.Selected Then
                    Select Case CameraMoveMode
                        Case CameraMovement.Left
                            mdl.Matrix.Xfrom1 += moveObjectBy
                        Case CameraMovement.LeftRight
                            mdl.Matrix.Yfrom1 += moveObjectBy
                        Case CameraMovement.Right
                            mdl.Matrix.Zfrom1 += moveObjectBy
                    End Select
                End If
            Next
            For Each lit In Scene.Lights
                If lit.Selected Then
                    Select Case CameraMoveMode
                        Case CameraMovement.Left
                            lit.Location.X += moveObjectBy
                        Case CameraMovement.LeftRight
                            lit.Location.Y += moveObjectBy
                        Case CameraMovement.Right
                            lit.Location.Z += moveObjectBy
                    End Select
                End If
            Next
            For Each act In Scene.Actives
                If act.Selected Then
                    Select Case CameraMoveMode
                        Case CameraMovement.Left
                            act.Location.X += moveObjectBy
                        Case CameraMovement.LeftRight
                            act.Location.Y += moveObjectBy
                        Case CameraMovement.Right
                            act.Location.Z += moveObjectBy
                    End Select
                End If
            Next
            For Each wnd In Scene.Windows
                If wnd.Selected Then
                    Dim delta As Vector
                    Select Case CameraMoveMode
                        Case CameraMovement.Left
                            delta.X += moveObjectBy
                        Case CameraMovement.LeftRight
                            delta.Y += moveObjectBy
                        Case CameraMovement.Right
                            delta.Z += moveObjectBy
                    End Select

                    For iV = 0 To wnd.Verts.Count - 1
                        wnd.Verts(iV) += delta
                    Next
                End If
            Next

            UpdateView()
        Else ' Moving camera
            Select Case CameraMoveMode
                Case CameraMovement.Right
                    myGLView.CameraRotationH += MouseFrameDeltaX * RotationSpeed
                    myGLView.CameraRotationV += MouseFrameDeltaY * RotationSpeed
                    myGLView.CameraRotationV = Math.Max(-90, Math.Min(90, myGLView.CameraRotationV))

                    UpdateView()
                Case CameraMovement.Left
                    myGLView.CameraRotationH += MouseFrameDeltaX * RotationSpeed

                    Dim moveDelta As Vector
                    moveDelta.X = Math.Cos(DegToRad(-myGLView.CameraRotationH + 90))
                    moveDelta.Z = Math.Sin(DegToRad(-myGLView.CameraRotationH + 90))
                    moveDelta *= -MouseFrameDeltaY
                    moveDelta *= MoveSpeed

                    myGLView.CameraPosition += moveDelta
                    UpdateView()
                Case CameraMovement.LeftRight
                    myGLView.CameraPosition.Y -= MouseFrameDeltaY * MoveSpeed

                    Dim moveDelta As Vector
                    moveDelta.X = Math.Cos(DegToRad(-myGLView.CameraRotationH))
                    moveDelta.Z = Math.Sin(DegToRad(-myGLView.CameraRotationH))
                    moveDelta *= MouseFrameDeltaX
                    moveDelta *= MoveSpeed

                    myGLView.CameraPosition += moveDelta

                    UpdateView()
            End Select
        End If

        PrevMouseX = e.X
        PrevMouseY = e.Y
    End Sub
#End Region

    Private Sub UpdateView()
        myGLView.GLDraw()
        myGLView.Refresh()
    End Sub

    Private Sub DeselectAll()
        For Each a In Project.Assets
            a.Selected = False
        Next
    End Sub

    Private Sub Preview_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        Scene = Project.Assets.Find(Function(x) x.Type = AssetType.Scene)
        If Scene Is Nothing Then
            MessageBox.Show("This project doesn't have a scene.")
            Close()
        End If

        cmbGrid.SelectedIndex = 4

        apRoom.Project = Project
        apWinRoom1.Project = Project
        apWinRoom2.Project = Project
        apRoom.LoadData()
        apWinRoom1.LoadData()
        apWinRoom2.LoadData()
        apRoom.SelectedAsset = Nothing

        cmsAddDefaultActive.Items.Clear()
        Dim defaultActiveNames As New List(Of String)
        For Each desc In Project.ActiveDescription.Chunks
            defaultActiveNames.Add(desc.Name)
        Next
        defaultActiveNames.Sort()
        For Each act In defaultActiveNames
            cmsAddDefaultActive.Items.Add(act)
        Next

        myGLView = New GLViewControl
        myGLView.Project = Project
        myGLView.Dock = DockStyle.Fill
        myGLView.Parent = ThreeDPanel
        myGLView.AllowDrop = True
    End Sub

    Private Sub ThreeDViewForm_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        myGLView.Dispose()
    End Sub

    Property CurrentRoom As Room
        Get
            Return apRoom.SelectedAsset
        End Get
        Set(value As Room)
            apRoom.SelectedAsset = value
            SetRoom(value)
        End Set
    End Property

    Private Sub llHelp_LinkClicked(sender As System.Object, e As System.Windows.Forms.LinkLabelLinkClickedEventArgs) Handles llHelp.LinkClicked
        MessageBox.Show(My.Resources.ThreeDHelp, "3D edit help")
    End Sub

    Private Sub myGLView_DragEnter(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles myGLView.DragEnter
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim files() As String = e.Data.GetData(DataFormats.FileDrop)

            Select Case LCase(IO.Path.GetExtension(files(0)))
                Case ".obj", ".bmp", ".jpg", ".jpeg", ".png", ".gif", ".tif", ".tiff", ".wmf", ".emf", ".ico"
                    e.Effect = DragDropEffects.All
            End Select
        End If
    End Sub

    Private Sub myGLView_DragDrop(sender As System.Object, e As System.Windows.Forms.DragEventArgs) Handles myGLView.DragDrop
        If e.Data.GetDataPresent(DataFormats.FileDrop) Then
            Dim files() As String = e.Data.GetData(DataFormats.FileDrop)
            Select Case LCase(IO.Path.GetExtension(files(0)))
                Case ".obj"
                    AddOBJ(files(0))
                Case ".bmp", ".jpg", ".jpeg", ".png", ".gif", ".tif", ".tiff", ".wmf", ".emf", ".ico"
                    Dim bmp = New Bitmap(files(0))
                    If IsPowerOfTwo(bmp.Width) And IsPowerOfTwo(bmp.Height) Then
                        Dim newTexAsset As Texture = Project.CreateAsset(AssetType.Texture)
                        newTexAsset.Name = Project.EnsureUniqueName(IO.Path.GetFileNameWithoutExtension(files(0)))
                        newTexAsset.Bitmap = bmp
                        MessageBox.Show("Added texture $" & newTexAsset.ID & " " & newTexAsset.Name)
                    Else
                        MessageBox.Show("Width and height must be power of two.")
                    End If
            End Select
        End If
    End Sub

    Sub AddOBJ(file As String)
        If CurrentRoom Is Nothing Then
            MessageBox.Show("Select a room first.")
        Else
            Dim newModel As Model = Project.CreateAsset(AssetType.Model)
            newModel.Name = Project.EnsureUniqueName(IO.Path.GetFileNameWithoutExtension(file))
            newModel.Room = CurrentRoom
            newModel.Matrix.Xfrom1 = (myGLView.CameraPosition.X \ Grid) * Grid
            newModel.Matrix.Yfrom1 = (myGLView.CameraPosition.Y \ Grid) * Grid
            newModel.Matrix.Zfrom1 = (myGLView.CameraPosition.Z \ Grid) * Grid
            newModel.LoadOBJ(file)
            If cbAddInScene.Checked Then
                Scene.Models.Add(newModel)
            End If

            UpdateView()

            MessageBox.Show("Added model $" & newModel.ID & " " & newModel.Name)
        End If
    End Sub

    Private Sub btnProperties_Click(sender As System.Object, e As System.EventArgs) Handles btnProperties.Click
        For Each a In Project.Assets
            If a.Selected Then
                Select Case a.Type
                    Case AssetType.Model
                        ModelProps.Model = a
                        ModelProps.ShowDialog()
                        Return
                    Case AssetType.Light
                        LightProps.Light = a
                        LightProps.ShowDialog()
                        Return
                    Case AssetType.ActiveA, AssetType.ActiveB
                        Dim propsForm As New ActiveProps
                        propsForm.Active = a
                        propsForm.ShowDialog(Me)
                        Return
                End Select
            End If
        Next
    End Sub

    Private Sub btnAddModel_Click(sender As System.Object, e As System.EventArgs) Handles btnAddModel.Click
        If OpenObjDlg.ShowDialog = Windows.Forms.DialogResult.OK Then
            AddOBJ(OpenObjDlg.FileName)
        End If
    End Sub

    Private Sub btnAddLight_Click(sender As System.Object, e As System.EventArgs) Handles btnAddLight.Click
        If CurrentRoom Is Nothing Then
            MessageBox.Show("Select a room first.")
        Else
            Dim newLight As Light = Project.CreateAsset(AssetType.Light)
            newLight.Room = CurrentRoom
            newLight.Location.X = (myGLView.CameraPosition.X \ Grid) * Grid
            newLight.Location.Y = (myGLView.CameraPosition.Y \ Grid) * Grid
            newLight.Location.Z = (myGLView.CameraPosition.Z \ Grid) * Grid
            newLight.Name = Project.EnsureUniqueName("Light")

            If cbAddInScene.Checked Then
                Scene.Lights.Add(newLight)
            End If

            UpdateView()
        End If

    End Sub

    Private Sub btnAddWindow_Click(sender As System.Object, e As System.EventArgs) Handles btnAddWindow.Click
        Dim newWindow As Window = Project.CreateAsset(AssetType.Window)
        newWindow.Name = Project.EnsureUniqueName("Room_window")
        newWindow.Room1 = apWinRoom1.SelectedAsset
        newWindow.Room2 = apWinRoom2.SelectedAsset

        Dim origin As Vector
        origin.X = (myGLView.CameraPosition.X \ Grid) * Grid
        origin.Y = (myGLView.CameraPosition.Y \ Grid) * Grid
        origin.Z = (myGLView.CameraPosition.Z \ Grid) * Grid

        newWindow.Verts.Add(origin + New Vector(-128, -128, 0))
        newWindow.Verts.Add(origin + New Vector(-128, 128, 0))
        newWindow.Verts.Add(origin + New Vector(128, 128, 0))
        newWindow.Verts.Add(origin + New Vector(128, -128, 0))
        Scene.Windows.Add(newWindow)

        UpdateView()
    End Sub

    Private Sub btnAddRoom_Click(sender As System.Object, e As System.EventArgs) Handles btnAddRoom.Click
        Dim name = InputBox("New room name:")
        If name <> "" Then
            Dim newRoom = Project.CreateAsset(AssetType.Room)
            newRoom.Name = Project.EnsureUniqueName(name)
            Scene.Rooms.Add(newRoom)

            apRoom.LoadData()
            apWinRoom1.LoadData()
            apWinRoom2.LoadData()
        End If
    End Sub

    Private Sub btnActives_Click(sender As System.Object, e As System.EventArgs) Handles btnActives.Click
        ActiveTreeForm.Project = Project
        ActiveTreeForm.ShowDialog(Me)
    End Sub

    Private Sub tbSearch_TextChanged(sender As System.Object, e As System.EventArgs) Handles tbSearch.TextChanged
        If tbSearch.Text <> "" Then
            Dim text = tbSearch.Text
            If text.StartsWith("$") Then text = text.Substring(1)
            Dim id = -1
            If Integer.TryParse(text, id) Then
                For Each mdl In Scene.Models
                    If mdl.ID = id Then
                        GoToEntity(mdl)
                        Return
                    End If
                Next
                For Each lit In Scene.Lights
                    If lit.ID = id Then
                        GoToEntity(lit)
                        Return
                    End If
                Next
                For Each act In Scene.Actives
                    If act.ID = id Then
                        GoToEntity(act)
                        Return
                    End If
                Next
            Else
                For Each mdl In Scene.Models
                    If mdl.Name = text Then
                        GoToEntity(mdl)
                        Return
                    End If
                Next
                For Each lit In Scene.Lights
                    If lit.Name = text Then
                        GoToEntity(lit)
                        Return
                    End If
                Next
                For Each act In Scene.Actives
                    If act.Name = text Then
                        GoToEntity(act)
                        Return
                    End If
                Next
            End If
        End If
    End Sub

    Sub GoToEntity(ent As Asset)
        For Each mdl In Scene.Models
            mdl.Selected = False
        Next
        For Each lit In Scene.Lights
            lit.Selected = False
        Next
        For Each act In Scene.Actives
            act.Selected = False
        Next

        ent.Selected = True

        myGLView.CameraPosition = ent.GetLocation()
        myGLView.CameraPosition.Z -= 128
        myGLView.CameraRotationH = 0
        myGLView.CameraRotationV = 0
        UpdateView()
    End Sub

    Sub OnWindowSelectionChange()
        Dim selectedWindow = Scene.Windows.Find(Function(x) x.Selected)
        If selectedWindow IsNot Nothing Then
            apWinRoom1.SelectedAsset = selectedWindow.Room1
            apWinRoom2.SelectedAsset = selectedWindow.Room2
        Else
            apWinRoom1.SelectedAsset = Nothing
            apWinRoom2.SelectedAsset = Nothing
        End If
    End Sub

    Private Sub rbModeWindows_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles rbModeWindows.CheckedChanged
        myGLView.ShowWindows = rbModeWindows.Checked
        gbWindow.Enabled = rbModeWindows.Checked
        btnAddWindow.Enabled = rbModeWindows.Checked
        apRoom.Enabled = Not rbModeWindows.Checked
        If rbModeWindows.Checked Then
            CurrentRoom = Nothing
        End If
        DeselectAll()
        UpdateView()
    End Sub

    Private Sub TransformStuff(mtx As Matrix)
        For Each mdl In Scene.Models
            If mdl.Selected Then
                mdl.Matrix *= mtx
            End If
        Next
        For Each wnd In Scene.Windows
            If wnd.Selected Then
                Dim origin = wnd.GetLocation()

                For iV = 0 To wnd.Verts.Count - 1
                    wnd.Verts(iV) = (wnd.Verts(iV) - origin).TransformBy(mtx) + origin
                Next
            End If
        Next
        UpdateView()
    End Sub

    Private Sub btnScaleMult_Click(sender As System.Object, e As System.EventArgs) Handles btnScaleMult.Click
        TransformStuff(Matrix.Scale(New Vector(nuScaleX.Value, nuScaleY.Value, nuScaleZ.Value)))
    End Sub

    Private Sub btnScaleDiv_Click(sender As System.Object, e As System.EventArgs) Handles btnScaleDiv.Click
        TransformStuff(Matrix.Scale(New Vector(1 / nuScaleX.Value, 1 / nuScaleY.Value, 1 / nuScaleZ.Value)))
    End Sub

    Private Sub apWinRoom1_UserPickedAsset(sender As System.Object, e As System.EventArgs) Handles apWinRoom1.UserPickedAsset
        Dim selectedWindow = Scene.Windows.Find(Function(x) x.Selected)
        If selectedWindow IsNot Nothing Then
            selectedWindow.Room1 = apWinRoom1.SelectedAsset
        End If
    End Sub

    Private Sub apWinRoom2_UserPickedAsset(sender As System.Object, e As System.EventArgs) Handles apWinRoom2.UserPickedAsset
        Dim selectedWindow = Scene.Windows.Find(Function(x) x.Selected)
        If selectedWindow IsNot Nothing Then
            selectedWindow.Room2 = apWinRoom2.SelectedAsset
        End If
    End Sub

    Private Sub SetRoom(rm As Room)
        If rm Is Nothing Then
            For Each a In Project.Assets
                a.Visible = True
            Next
        Else
            For Each a In Project.Assets
                a.Visible = (a.Room Is rm)
            Next
        End If

        Refresh()
    End Sub

    Private Sub apRoom_UserPickedAsset(sender As System.Object, e As System.EventArgs) Handles apRoom.UserPickedAsset
        SetRoom(CurrentRoom)
    End Sub

    Private Sub btnAddDefaultActive_Click(sender As System.Object, e As System.EventArgs) Handles btnAddDefaultActive.Click
        If CurrentRoom Is Nothing Then
            MessageBox.Show("Select a room first.")
        Else
            cmsAddDefaultActive.Show(MousePosition)
        End If
    End Sub

    Private Sub cmsAddDefaultActive_ItemClicked(sender As System.Object, e As System.Windows.Forms.ToolStripItemClickedEventArgs) Handles cmsAddDefaultActive.ItemClicked
        Dim desc = Project.ActiveDescription.Chunks.Find(Function(x) x.Name = e.ClickedItem.Text)
        If desc IsNot Nothing Then
            Dim newActive As ActiveA = Project.CreateAsset(AssetType.ActiveA)
            newActive.Name = Project.EnsureUniqueName(desc.Name)
            newActive.Room = CurrentRoom
            newActive.Location.X = (myGLView.CameraPosition.X \ Grid) * Grid
            newActive.Location.Y = (myGLView.CameraPosition.Y \ Grid) * Grid
            newActive.Location.Z = (myGLView.CameraPosition.Z \ Grid) * Grid

            newActive.Describe = desc
            For Each def In desc.ParameterLines
                Dim actParam As New ActiveParam
                actParam.Definition = def
                Select Case def.Type
                    Case ParamType.PT_Flag, ParamType.PT_Float, ParamType.PT_List
                        If def.Misc <> "" Then
                            actParam.Value = def.Misc '???
                        End If
                End Select
                newActive.Params.Add(actParam)
            Next

            Scene.Actives.Add(newActive)

            UpdateView()
        End If
    End Sub

    Private Sub cbShowNonSceneEntities_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbShowNonSceneEntities.CheckedChanged
        UpdateView()
    End Sub

    Private Sub cbShowChildActives_CheckedChanged(sender As System.Object, e As System.EventArgs) Handles cbShowChildActives.CheckedChanged
        myGLView.ShowChildActives = cbShowChildActives.Checked
        UpdateView()
    End Sub

    Private Sub labRotPlusX_Click(sender As System.Object, e As System.EventArgs) Handles labRotPlusX.Click, labRotPlusX.DoubleClick
        TransformStuff(Matrix.RotationX(nuRotAngle.Value))
    End Sub

    Private Sub labRotMinusX_Click(sender As System.Object, e As System.EventArgs) Handles labRotMinusX.Click, labRotMinusX.DoubleClick
        TransformStuff(Matrix.RotationX(-nuRotAngle.Value))
    End Sub

    Private Sub labRotPlusY_Click(sender As System.Object, e As System.EventArgs) Handles labRotPlusY.Click, labRotPlusY.DoubleClick
        TransformStuff(Matrix.RotationY(nuRotAngle.Value))
    End Sub

    Private Sub labRotMinusY_Click(sender As System.Object, e As System.EventArgs) Handles labRotMinusY.Click, labRotMinusY.DoubleClick
        TransformStuff(Matrix.RotationY(-nuRotAngle.Value))
    End Sub

    Private Sub labRotPlusZ_Click(sender As System.Object, e As System.EventArgs) Handles labRotPlusZ.Click, labRotPlusZ.DoubleClick
        TransformStuff(Matrix.RotationZ(nuRotAngle.Value))
    End Sub

    Private Sub labRotMinusZ_Click(sender As System.Object, e As System.EventArgs) Handles labRotMinusZ.Click, labRotMinusZ.DoubleClick
        TransformStuff(Matrix.RotationZ(-nuRotAngle.Value))
    End Sub

    Private Sub cmbGrid_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles cmbGrid.SelectedIndexChanged
        If cmbGrid.SelectedItem IsNot Nothing Then
            Grid = cmbGrid.SelectedItem
        End If
    End Sub

    Private Sub btnDuplicate_Click(sender As System.Object, e As System.EventArgs) Handles btnDuplicate.Click
        Dim copies As New List(Of Asset)
        Dim showWarning As Boolean
        For Each a In Project.Assets
            If a.Selected Then
                Select Case a.Type
                    Case AssetType.Model : copies.Add(DirectCast(a, Model).Clone())
                    Case AssetType.Light : copies.Add(DirectCast(a, Light).Clone())
                    Case Else : showWarning = True
                End Select
            End If
        Next

        DeselectAll()

        For Each copy In copies
            Project.AllocateID(copy)
            copy.Name = Project.EnsureUniqueName(copy.Name)
            Project.Assets.Add(copy)
            copy.AddLocation(New Vector(Grid, 0, Grid))

            Select Case copy.Type
                Case AssetType.Model : Scene.Models.Add(copy)
                Case AssetType.Light : Scene.Lights.Add(copy)
            End Select
        Next

        If showWarning Then
            MessageBox.Show("You can only duplicate models and lights for now.")
        End If

        UpdateView()

    End Sub

    Private Sub btnDelete_Click(sender As System.Object, e As System.EventArgs) Handles btnDelete.Click
        If MessageBox.Show("Delete selected entities?" & vbCrLf & vbCrLf & "Be careful, this may make the level unopenable.", "Delete", MessageBoxButtons.YesNo) = Windows.Forms.DialogResult.Yes Then
            Scene.Models.RemoveAll(Function(x) x.Selected)
            Scene.Lights.RemoveAll(Function(x) x.Selected)
            Scene.Windows.RemoveAll(Function(x) x.Selected)
            Scene.Actives.RemoveAll(Function(x) x.Selected)

            Project.RemoveAssets(Function(x) x.Selected)

            UpdateView()
        End If
    End Sub

    Private Sub btnDeleteRoom_Click(sender As System.Object, e As System.EventArgs) Handles btnDeleteRoom.Click
        Dim curRoom = CurrentRoom
        If curRoom Is Nothing Then Return

        For Each a In Project.Assets
            If a.Room Is curRoom Then
                MessageBox.Show("Room must be empty.")
                Return
            End If
        Next

        Scene.Rooms.Remove(curRoom)
        Project.RemoveAssets(Function(x) x Is curRoom)

        apRoom.LoadData()
        apWinRoom1.LoadData()
        apWinRoom2.LoadData()

        CurrentRoom = Nothing
    End Sub
End Class

Public Class GLViewControl
    Inherits OpenGLControl

    Public Project As Project

    Public Textures As New Dictionary(Of Integer, OpenGLTexture2D)
    Dim S_Light, S_LightOffset, S_LightCycle, S_ActiveLight, S_LensFlare,
        S_Active, S_Mover, S_Item, S_Player, S_Enemy,
        S_Object, S_Composite, S_Simple, S_Autostart,
        S_PathDelay, S_PathMove, S_PathRotate, S_PathStart, S_PathBreak,
        S_Platform, S_PlatformRotate, S_PlatformOffset, S_Rotator As OpenGLTexture2D

    Public CameraPosition As Vector = New Vector(256, 256, 256)
    Public CameraRotationH As Single
    Public CameraRotationV As Single
    Public ShowWindows As Boolean
    Public ShowChildActives As Boolean

    Public IsSelecting As Boolean 'Is this a hittest render?

    Private Sub SetTexture(tex As Texture)
        If tex Is Nothing Then
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        End If

        If Textures.ContainsKey(tex.ID) Then
            Textures(tex.ID).Bind()
        Else
            ' Create a new texture...
            Dim glTex = New OpenGLTexture2D(tex.Bitmap, OpenGLTexture2D.Tex2DType.MIPMAPED)
            glTex.Bind()
            OpenGL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR_MIPMAP_LINEAR)
            OpenGL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
            Textures.Add(tex.ID, glTex)
        End If
    End Sub

    Protected Overrides Sub initGLContext() 'Need to override this method
        GL.glEnable(GLFlags.GL_CULL_FACE) ' dont draw back
        GL.glShadeModel(GLFlags.GL_SMOOTH) 'Smooth Shading
        GL.glClearColor(0.0F, 0.0F, 0.0, 0.5F)               'Black Background
        GL.glClearDepth(1.0F)                                'Depth Buffer Setup
        GL.glEnable(GLFlags.GL_DEPTH_TEST) 'Enable Depth Testing
        GL.glDepthFunc(GLFlags.GL_LEQUAL)  'Type of Depth Testing to do
        'does this do anything?
        GL.glHint(GLFlags.GL_PERSPECTIVE_CORRECTION_HINT, GLFlags.GL_NICEST) 'Nice Perspective Calculations

        S_Light = New OpenGLAlphaTexture2D(My.Resources.S_Light)
        S_LightOffset = New OpenGLAlphaTexture2D(My.Resources.S_LightOffset)
        S_LightCycle = New OpenGLAlphaTexture2D(My.Resources.S_LightCycle)
        S_ActiveLight = New OpenGLAlphaTexture2D(My.Resources.S_ActiveLight)
        S_LensFlare = New OpenGLAlphaTexture2D(My.Resources.S_LensFlare)
        S_Active = New OpenGLAlphaTexture2D(My.Resources.S_Active)
        S_Enemy = New OpenGLAlphaTexture2D(My.Resources.S_Enemy)
        S_Item = New OpenGLAlphaTexture2D(My.Resources.S_Item)
        S_Mover = New OpenGLAlphaTexture2D(My.Resources.S_Mover)
        S_Player = New OpenGLAlphaTexture2D(My.Resources.S_Player)
        S_Object = New OpenGLAlphaTexture2D(My.Resources.S_Object)
        S_Composite = New OpenGLAlphaTexture2D(My.Resources.S_Composite)
        S_Simple = New OpenGLAlphaTexture2D(My.Resources.S_Simple)
        S_Autostart = New OpenGLAlphaTexture2D(My.Resources.S_Autostart)
        S_PathDelay = New OpenGLAlphaTexture2D(My.Resources.S_PathDelay)
        S_PathMove = New OpenGLAlphaTexture2D(My.Resources.S_PathMove)
        S_PathRotate = New OpenGLAlphaTexture2D(My.Resources.S_PathRotate)
        S_PathStart = New OpenGLAlphaTexture2D(My.Resources.S_PathStart)
        S_PathBreak = New OpenGLAlphaTexture2D(My.Resources.S_PathBreak)
        S_Platform = New OpenGLAlphaTexture2D(My.Resources.S_Platform)
        S_PlatformRotate = New OpenGLAlphaTexture2D(My.Resources.S_PlatformRotate)
        S_PlatformOffset = New OpenGLAlphaTexture2D(My.Resources.S_PlatformOffset)
        S_Rotator = New OpenGLAlphaTexture2D(My.Resources.S_Rotator)
    End Sub

    Protected Overrides Sub OnSizeChanged(ByVal e As EventArgs)
        MyBase.OnSizeChanged(e)
        Dim S As Size = Me.Size

        GL.glMatrixMode(GLFlags.GL_PROJECTION)
        GL.glLoadIdentity()
        GL.gluPerspective(90, S.Width / S.Height, 1, 65536.0)
        GL.glMatrixMode(GLFlags.GL_MODELVIEW)
        GL.glLoadIdentity()
    End Sub

    Private Sub DrawLine3D(x1 As Single, y1 As Single, z1 As Single, x2 As Single, y2 As Single, z2 As Single, col As Color)
        GL.glBegin(GLFlags.GL_LINES)
        GL.glColor3ub(col.R, col.G, col.B)
        GL.glVertex3f(x1, y1, z1)
        GL.glVertex3f(x2, y2, z2)
        GL.glEnd()
    End Sub

    Private Sub DrawLine3D(v1 As Vector, v2 As Vector, col As Color)
        DrawLine3D(v1.X, v1.Y, v1.Z, v2.X, v2.Y, v2.Z, col)
    End Sub

    Private Sub DrawRadius(rad As Single)
        Const numSegments = 32
        Dim mult = 1 / numSegments * 2 * Math.PI
        GL.glBegin(GL.GL_LINE_LOOP)
        For i = 0 To numSegments
            GL.glVertex3f(Math.Cos(i * mult) * rad, Math.Sin(i * mult) * rad, 0)
        Next
        GL.glEnd()
        GL.glBegin(GL.GL_LINE_LOOP)
        For i = 0 To numSegments
            GL.glVertex3f(Math.Cos(i * mult) * rad, 0, Math.Sin(i * mult) * rad)
        Next
        GL.glEnd()
        GL.glBegin(GL.GL_LINE_LOOP)
        For i = 0 To numSegments
            GL.glVertex3f(0, Math.Cos(i * mult) * rad, Math.Sin(i * mult) * rad)
        Next
        GL.glEnd()
    End Sub

    Sub DrawSprite(size As Single)
        GL.glPushMatrix()
        GL.glRotatef(CameraRotationH, 0.0, 1.0, 0.0)
        GL.glRotatef(CameraRotationV, 1.0, 0.0, 0.0)

        GL.glBegin(GL.GL_QUADS)
        GL.glTexCoord2f(0, 0)
        GL.glVertex3f(-size, -size, 0)
        GL.glTexCoord2f(0, 1)
        GL.glVertex3f(-size, size, 0)
        GL.glTexCoord2f(1, 1)
        GL.glVertex3f(size, size, 0)
        GL.glTexCoord2f(1, 0)
        GL.glVertex3f(size, -size, 0)
        GL.glEnd()

        GL.glPopMatrix()
    End Sub

    Private Sub DrawBackgroundGrid()
        Dim normal = Color.FromArgb(64, 64, 64)
        Dim dark = Color.FromArgb(48, 48, 48)

        Dim a, b As Vector

        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        GL.glColor3b(32, 32, 32)
        GL.glutWireCube(65536)

        For i = -32768 To 32768 Step 1024
            a.X = i
            b.X = i
            a.Z = -32768
            b.Z = 32768
            DrawLine3D(a, b, If(i = 0, normal, dark))
            a.X = -32768
            b.X = 32768
            a.Z = i
            b.Z = i
            DrawLine3D(a, b, If(i = 0, normal, dark))
        Next
    End Sub

    Private Sub SetSelectionCode(code As Integer)
        GL.glColor3ub(code And 255, (code >> 8) And 255, (code >> 16 And 255))
    End Sub

    Private Sub DrawModel(mdl As Model)
        If Not mdl.Visible Then Return

        GL.glPushMatrix()

        GL.glMultMatrixf({mdl.Matrix.XfromX, mdl.Matrix.YfromX, mdl.Matrix.ZfromX, 0,
                          mdl.Matrix.XfromY, mdl.Matrix.YfromY, mdl.Matrix.ZfromY, 0,
                          mdl.Matrix.XfromZ, mdl.Matrix.YfromZ, mdl.Matrix.ZfromZ, 0,
                          mdl.Matrix.Xfrom1, mdl.Matrix.Yfrom1, mdl.Matrix.Zfrom1, 1})

        If IsSelecting Then
            GL.glBegin(GL.GL_QUADS)
            SetSelectionCode(mdl.ID + 1)
            For i = 0 To mdl.Polygons.Count - 1
                Dim p = mdl.Polygons(i)

                For iV = 1 To 4
                    Dim v = mdl.Verts(p.V(iV))
                    GL.glVertex3f(v.X, v.Y, v.Z)
                Next
            Next
            GL.glEnd()
        Else
            For i = 0 To mdl.Polygons.Count - 1
                Dim p = mdl.Polygons(i)

                If i = 0 OrElse p.iTexture <> mdl.Polygons(i - 1).iTexture Then
                    If p.iTexture >= 0 Then
                        SetTexture(mdl.Textures(p.iTexture))
                    Else
                        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
                    End If
                    GL.glBegin(GL.GL_QUADS)
                End If

                If mdl.Selected Then
                    GL.glColor3f(0, 1, 0)
                Else
                    GL.glColor3f(p.Diffuse / 2 + p.EmissiveColor.X, p.Diffuse / 2 + p.EmissiveColor.Y, p.Diffuse / 2 + p.EmissiveColor.Z)
                End If

                For iV = 1 To 4
                    GL.glTexCoord2f(p.UV(iV).U, 1 - p.UV(iV).V)
                    Dim v = mdl.Verts(p.V(iV))
                    GL.glVertex3f(v.X, v.Y, v.Z)
                Next

                If i = mdl.Polygons.Count - 1 OrElse p.iTexture <> mdl.Polygons(i + 1).iTexture Then
                    GL.glEnd()
                End If
            Next
        End If

        GL.glPopMatrix()

        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
    End Sub

    Private Sub DrawLight(lit As Light)
        If Not lit.Visible Then Return
        If ShowWindows Then Return

        If IsSelecting Then
            SetSelectionCode(lit.ID + 1)
        Else
            S_Light.Bind()
            If lit.Selected Then
                GL.glColor3f(0, 1, 0)
            Else
                GL.glColor3f(0.1 + lit.Color.X / 256, 0.1 + lit.Color.Y / 256, 0.1 + lit.Color.Z / 256)
            End If
        End If

        GL.glPushMatrix()
        GL.glTranslatef(lit.Location.X, lit.Location.Y, lit.Location.Z)

        ' Masked
        GL.glAlphaFunc(GL.GL_GREATER, 0.5)
        GL.glEnable(GL.GL_ALPHA_TEST)

        DrawSprite(32)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        If Not IsSelecting And lit.Selected Then
            GL.glColor3f(lit.Color.X / 256, lit.Color.Y / 256, lit.Color.Z / 256)
            DrawRadius(lit.Radius)
        End If

        GL.glPopMatrix()
    End Sub

    Private Sub DrawActive(act As Active)
        If Not act.Visible Then Return
        If ShowWindows Then Return
        If Not ShowChildActives And act.Parent IsNot Nothing Then Return

        GL.glColor3f(1, 1, 1)

        If IsSelecting Then
            SetSelectionCode(act.ID + 1)
            GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        Else
            If TypeOf act Is ActiveA Then
                Dim aa = DirectCast(act, ActiveA)
                Select Case aa.Describe.Name
                    Case "enemy" : S_Enemy.Bind()
                    Case "simple_light" : S_ActiveLight.Bind()
                    Case "upgrade" : S_Item.Bind()
                    Case Else : S_Simple.Bind()
                End Select
            Else
                Dim ab = DirectCast(act, ActiveB)
                Select Case ab.PrefabName
                    Case "autostart(on)" : S_Autostart.Bind()
                    Case "path_delay" : S_PathDelay.Bind()
                    Case "path_move" : S_PathMove.Bind()
                    Case "path_rotate" : S_PathRotate.Bind()
                    Case "path_start_on" : S_PathStart.Bind()
                    Case "path_break" : S_PathBreak.Bind()
                    Case "object" : S_Object.Bind()
                    Case "platform" : S_Platform.Bind()
                    Case "platform(rotate)" : S_PlatformRotate.Bind()
                    Case "platform_offset" : S_PlatformOffset.Bind()
                    Case "light_offset" : S_LightOffset.Bind()
                    Case "light(cycle)" : S_LightCycle.Bind()
                    Case "lens_flare" : S_LensFlare.Bind()
                    Case "rotator" : S_Rotator.Bind()
                    Case "mover" : S_Mover.Bind()
                    Case Else : S_Composite.Bind()
                End Select
            End If

            If act.Selected Then
                GL.glColor3f(0, 1, 0)
            End If
        End If

        GL.glPushMatrix()
        GL.glTranslatef(act.Location.X, act.Location.Y, act.Location.Z)
        DrawSprite(32)
        GL.glPopMatrix()
    End Sub

    Private Sub DrawWindow(wnd As Window)
        If Not ShowWindows Then Return

        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        If Not IsSelecting Then
            If wnd.Selected Then
                GL.glColor3f(0, 1, 0)
            Else
                GL.glColor3f(0, 0.5, 0)
            End If
            GL.glBegin(GL.GL_LINE_LOOP)
        Else
            SetSelectionCode(wnd.ID + 1)
            GL.glDisable(GL.GL_CULL_FACE)
            GL.glBegin(GL.GL_TRIANGLE_FAN)
        End If

        For Each vert In wnd.Verts
            GL.glVertex3f(vert.X, vert.Y, vert.Z)
        Next

        GL.glEnd()

        If IsSelecting Then
            GL.glEnable(GL.GL_CULL_FACE)
        Else
            If wnd.Selected Then
                'Draw direction
                Dim origin = wnd.GetLocation()
                Dim normal = Vector.Cross(wnd.Verts(1) - wnd.Verts(0), wnd.Verts(2) - wnd.Verts(0)).Normalize()
                GL.glBegin(GL.GL_LINES)
                GL.glColor3f(1, 0, 0)
                GL.glVertex3f(origin.X, origin.Y, origin.Z)
                GL.glVertex3f(origin.X + normal.X * 64, origin.Y + normal.Y * 64, origin.Z + normal.Z * 64)
                GL.glColor3f(0, 0, 1)
                GL.glVertex3f(origin.X, origin.Y, origin.Z)
                GL.glVertex3f(origin.X - normal.X * 64, origin.Y - normal.Y * 64, origin.Z - normal.Z * 64)
                GL.glEnd()
            End If
        End If

    End Sub

    Public Overrides Sub GLDraw() 'We use this method to draw our objects on screen
        GL.glDisable(GL.GL_BLEND)
        GL.glDepthMask(GL.GL_TRUE)

        If IsSelecting Then
            GL.glDisable(GL.GL_DITHER)
            GL.glDisable(GL.GL_FOG)
            GL.glDisable(GL.GL_LIGHTING)
            GL.glDisable(GL.GL_TEXTURE_1D)
            GL.glDisable(GL.GL_TEXTURE_2D)
            GL.glDisable(GL.GL_TEXTURE_3D)
            GL.glShadeModel(GL.GL_FLAT)
        Else
            GL.glEnable(GL.GL_TEXTURE_2D)
        End If

        GL.glClearColor(0, 0, 0, 1)
        GL.glClear(GLFlags.GL_COLOR_BUFFER_BIT Or GLFlags.GL_DEPTH_BUFFER_BIT)

        GL.glMatrixMode(GLFlags.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMultMatrixf({1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1}) 'remap coord system
        GL.glRotatef(-CameraRotationV, 1, 0, 0) 'up/down
        GL.glRotatef(-CameraRotationH, 0, 1, 0) 'left/right
        GL.glTranslatef(-CameraPosition.X, -CameraPosition.Y, -CameraPosition.Z)
        GL.glFrontFace(GL.GL_CW) 'Yup, now we need that

        If Not IsSelecting Then
            DrawBackgroundGrid()
            GL.glClear(GLFlags.GL_DEPTH_BUFFER_BIT)
            GL.glEnable(GL.GL_TEXTURE_2D)
        End If

        If Not (IsSelecting And ShowWindows) Then
            If ThreeDEditorForm.cbShowNonSceneEntities.Checked Then
                For Each a In ThreeDEditorForm.Project.Assets
                    Select Case a.Type
                        Case AssetType.Model : DrawModel(a)
                        Case AssetType.Light : DrawLight(a)
                    End Select
                Next
            Else
                For Each mdl In ThreeDEditorForm.Scene.Models
                    DrawModel(mdl)
                Next
                For Each lit In ThreeDEditorForm.Scene.Lights
                    DrawLight(lit)
                Next
            End If
            For Each act In ThreeDEditorForm.Scene.Actives
                DrawActive(act)
            Next
        End If
        For Each wnd In ThreeDEditorForm.Scene.Windows
            DrawWindow(wnd)
        Next

        GL.glFlush()
    End Sub

    Function GetIDUnderPoint(pos As Point) As Integer
        Dim pixel(3) As Byte
        IsSelecting = True
        GLDraw()
        IsSelecting = False
        GL.glReadPixels(pos.X, Me.Height - pos.Y, 1, 1, GL.GL_RGB, GL.GL_UNSIGNED_BYTE, pixel)
        Return (CInt(pixel(0)) Or (CInt(pixel(1)) << 8) Or (CInt(pixel(2)) << 16)) - 1
    End Function

End Class