﻿Imports CsGL.OpenGL

Module Utils
    Function NumToStr(val As Single) As String
        Return val.ToString("F6", Globalization.CultureInfo.InvariantCulture)
    End Function

    Function StrToNum(val As String) As Single
        Return Single.Parse(val, Globalization.CultureInfo.InvariantCulture)
    End Function

    Function StrEq(str1 As String, str2 As String) As Boolean
        Return String.Equals(str1, str2, StringComparison.OrdinalIgnoreCase)
    End Function

    Function IsPowerOfTwo(num As Integer) As Boolean
        Dim i = 1
        While i <= 2 ^ 16
            If num = i Then
                Return True
            End If
            i <<= 1
        End While

        Return False
    End Function

    Function ReadFixedZString(r As IO.BinaryReader, length As Integer) As String
        Dim s As New System.Text.StringBuilder
        Dim origPos = r.BaseStream.Position
        Dim c As Byte

        For i = 0 To length - 1
            c = r.ReadByte()

            If c <> 0 Then
                s.Append(Chr(c))
            Else
                r.BaseStream.Position = origPos + length
                Exit For
            End If
        Next

        Return s.ToString
    End Function

    Sub WriteFixedString(w As IO.BinaryWriter, text As String, length As Integer)
        For i = 0 To length - 1
            If i < text.Length Then
                w.Write(CByte(Asc(text(i))))
            Else
                w.Write(CByte(0))
            End If
        Next
    End Sub

    Function ReadVector(r As IO.BinaryReader) As Vector
        Dim v As Vector
        v.X = r.ReadSingle()
        v.Y = r.ReadSingle()
        v.Z = r.ReadSingle()
        Return v
    End Function

    Sub WriteVector(w As IO.BinaryWriter, vect As Vector)
        w.Write(vect.X)
        w.Write(vect.Y)
        w.Write(vect.Z)
    End Sub

    Function ReadMatrix(r As IO.BinaryReader) As Matrix
        Dim matrix As New Matrix
        matrix.Xfrom1 = r.ReadSingle()
        matrix.Yfrom1 = r.ReadSingle()
        matrix.Zfrom1 = r.ReadSingle()
        matrix.XfromX = r.ReadSingle()
        matrix.YfromX = r.ReadSingle()
        matrix.ZfromX = r.ReadSingle()
        matrix.XfromY = r.ReadSingle()
        matrix.YfromY = r.ReadSingle()
        matrix.ZfromY = r.ReadSingle()
        matrix.XfromZ = r.ReadSingle()
        matrix.YfromZ = r.ReadSingle()
        matrix.ZfromZ = r.ReadSingle()
        Return matrix
    End Function

    Sub ListAddUnique(Of T)(lst As List(Of T), item As T)
        Dim index = lst.BinarySearch(item)
        If index < 0 Then
            lst.Insert(Not index, item)
        End If
    End Sub

    Class OpenGLAlphaTexture2D
        Inherits OpenGLTexture2D

        Sub New(img As Bitmap)
            MyBase.New(img)
        End Sub

        Protected Overrides Function [New]() As Object
            ' BGR -> RGB
            Dim image2 As New Bitmap(image.Width, image.Height, Imaging.PixelFormat.Format32bppArgb)
            Dim fastBmp As New FastPixel(image)
            Dim fastBmp2 As New FastPixel(image2)
            fastBmp.Lock()
            fastBmp2.Lock()
            For y = 0 To image.Height - 1
                For x = 0 To image.Width - 1
                    Dim pix = fastBmp.GetPixel(x, y)
                    fastBmp2.SetPixel(x, y, Color.FromArgb(pix.A, pix.B, pix.G, pix.R))
                Next
            Next
            fastBmp.Unlock(False)
            fastBmp2.Unlock(True)
            Me.image = image2

            Dim textures As UInteger() = {0}
            Dim bitmapdata = Me.image.LockBits(New Rectangle(0, 0, Me.image.Width, Me.image.Height), Imaging.ImageLockMode.ReadOnly, Imaging.PixelFormat.Format32bppArgb)
            GL.glGenTextures(textures.Length, textures)
            OpenGL.glBindTexture(3553, textures(0))
            Me.SetParameter()
            If (Me.IsMipmaped) Then
                GL.gluBuild2DMipmaps(3553, GL.GL_RGBA, Me.image.Width, Me.image.Height, GL.GL_RGBA, 5121, bitmapdata.Scan0)
            Else
                GL.glTexImage2D(3553, 0, GL.GL_RGBA, Me.image.Width, Me.image.Height, If(Me.IsBorder, 1, 0), GL.GL_RGBA, 5121, bitmapdata.Scan0)
            End If
            Me.image.UnlockBits(bitmapdata)
            OpenGLException.Assert()
            Return textures
        End Function

    End Class

End Module
